/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComScale.cpp
    \brief    Functions for upscale and downscale of pictures for dynamic resolution coding
*/

#include <TComScale.h>
#include <assert.h>

//! \ingroup TLibCommon
//! \{

/** Separable upscale 2x2 with samples at phases +1/4 and -1/4 with respect to low-res grid
 * \returns void
 */
void upScaleComponent2x2QuarterPhase( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){

  const int iDestHeight = iSrcHeight*2;

  const int NUM_UP_TAPS = 4;
  const int iTaps[NUM_UP_TAPS] = {  -4,  54,  16,   -2};
  const int iFilterShift = 6;
  const int NUM_UP_TAPS2 = NUM_UP_TAPS/2;

  // Vertical
  const int iTempStride = iSrcWidth+2*NUM_UP_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + NUM_UP_TAPS;
  int iTemp;

  for (int j=0; j<iSrcHeight; ++j){
    // -1/4 phase
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * pDataSrc[i - (k-NUM_UP_TAPS2+1)*iSrcStride];
      }
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );

    pDataDest += iTempStride;

    // +1/4 phase
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * pDataSrc[i + (k-NUM_UP_TAPS2+1)*iSrcStride];
      }
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );

    pDataDest += iTempStride;
    pDataSrc += iSrcStride;
  }

  pDataSrc = pTemp + NUM_UP_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iSrcWidth; ++i, ii+=2){
      // -1/4 phase
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * pDataSrc[i - (k-NUM_UP_TAPS2+1)];
      }
      pDataDest[ii] = Clip(iTemp>>iFilterShift);
      // +1/4 phase
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * pDataSrc[i + (k-NUM_UP_TAPS2+1)];
      }
      pDataDest[ii+1] = Clip(iTemp>>iFilterShift);
    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

static int up_filt_taps0[2][4] = { {16, 0, 0, 0},            /*f7*/
                                   {20, -4, 0, 0}            /*Linear*/};
static int up_filt_half_lens0[2] = {1, 2};
static int up_filt_shifts0[2] = {5, 5};

/** Separable upscale 2x2 with samples at phases 0 and +1/2 with respect to low-res grid
 * \returns void
 */
void upScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight, int filt_type){

  const int iDestHeight = iSrcHeight*2;
  const int NUM_UP_TAPS = up_filt_half_lens0[filt_type];
  const int* iTaps = up_filt_taps0[filt_type];
  const int iFilterShift = up_filt_shifts0[filt_type];

  const int iTempStride = iSrcWidth+2*NUM_UP_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];
  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + NUM_UP_TAPS;
  int iTemp;

  for (int j=0; j<iSrcHeight; ++j){
    // Zero phase
    for (int i=0; i<iSrcWidth; ++i){
      pDataDest[i] = pDataSrc[i];
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );
    pDataDest += iTempStride;

    // +1/2 phase
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i + (k+1)*iSrcStride]  + pDataSrc[i - k*iSrcStride] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - NUM_UP_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+NUM_UP_TAPS, pDataDest[iSrcWidth-1] );

    pDataDest += iTempStride;
    pDataSrc += iSrcStride;
  }

  pDataSrc = pTemp + NUM_UP_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iSrcWidth; ++i, ii+=2){
      pDataDest[ii] = pDataSrc[i];
      iTemp = 0;
      for (int k=0; k<NUM_UP_TAPS; ++k){
        iTemp += iTaps[k] * (pDataSrc[i - k] + pDataSrc[i + k+1] );
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[ii+1] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

/** Non-normative 2x2 upscale used e.g. for output only
 * \returns void
 */
void upScalePic(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==1 && iDestResIdx == 0){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    upScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight(), 1 );
    upScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 1 );
    upScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 1 );
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}

/** Normative 2x2 reference picture upscaling
 * \returns void
 */
void upScaleReference(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==1 && iDestResIdx == 0){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    if (REF_UP_FILT==0){
      upScaleComponent2x2QuarterPhase( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight() );
      upScaleComponent2x2QuarterPhase( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
      upScaleComponent2x2QuarterPhase( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2 );
    } else {
      upScaleComponent2x2( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight(),REF_UP_FILT-1 );
      upScaleComponent2x2( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, REF_UP_FILT-1 );
      upScaleComponent2x2( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, REF_UP_FILT-1 );
    }
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}

static int dn_filt_taps[4][6] =  { {27, 6, -1, 0, 0, 0},      /*Lanczos 6-tap*/
                                   {57, 18, -9, -5, 2, 1},    /*JSVM 2.0 window (sine-windowed sinc)*/
                                   {2, 0, 0, 0, 0, 0},        /*Linear */
				   {29, 3, 0, 0, 0, 0}        /*4-tap Lanczos*/};
static int dn_filt_half_lens[4] = {3, 6, 1, 2};
static int dn_filt_shifts[4] = {6, 7, 2, 6};

/** Downscale a component array by 2/2 with samples at phase +1/2 with respect to high-res grid
 * \returns void
 */
void downScaleComponent2x2HalfPhase( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight){

  const int iDestWidth = iSrcWidth/2;
  const int iDestHeight = iSrcHeight/2;

  const int HALF_NUM_DOWN_TAPS = dn_filt_half_lens[REF_DN_FILT-1];
  const int iFilterShift = dn_filt_shifts[REF_DN_FILT-1];
  const int* iTaps = dn_filt_taps[REF_DN_FILT-1];

  const int iTempStride = iDestWidth;
  Pel* pTemp = new Pel[ (iSrcHeight + 2*HALF_NUM_DOWN_TAPS) * iDestWidth];

  int iTemp;

  // Horizontal
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp+HALF_NUM_DOWN_TAPS*iTempStride;

  for (int j=0; j<iSrcHeight; ++j){
    for (int i=0,ii=0; i<iDestWidth; ++i,ii+=2){
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[ii+k+1] + pDataSrc[ii-k]);
      }
      pDataDest[i] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iSrcStride;
    pDataDest += iTempStride;
  }
  pDataDest = pTemp;
  Pel* pFirst = pTemp+HALF_NUM_DOWN_TAPS*iTempStride;
  for (int j=0; j<HALF_NUM_DOWN_TAPS; ++j){
    std::copy(pFirst, pFirst+iTempStride, pDataDest );
    pDataDest += iTempStride;
  }
  pDataDest = pTemp+(HALF_NUM_DOWN_TAPS+iSrcHeight)*iTempStride;
  pFirst = pDataDest-iTempStride;
  for (int j=0; j<HALF_NUM_DOWN_TAPS; ++j){
    std::copy(pFirst, pFirst+iTempStride, pDataDest );
    pDataDest += iTempStride;
  }

  // Vertical
  pDataSrc = pTemp + HALF_NUM_DOWN_TAPS*iTempStride;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0; i<iDestWidth; ++i){
      iTemp = 1<<(iFilterShift-1);
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[i+(k+1)*iTempStride] + pDataSrc[i-k*iTempStride]);
      }
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    pDataSrc += 2*iTempStride;
    pDataDest += iDestStride;
  }


  delete[] pTemp;

}

static int dn_filt_taps0[2][12] = { {36, 0, -4, 0, 0, 0}, /* 3-lobe Lanczos*/
                                    {40, 0, -12, 0, 4, 0},/*h11*/
                                  };
static int dn_filt_half_lens0[2] = {3, 5};
static int dn_filt_shifts0[2] = {7, 7};
/** Downscale a component array by 2/2 with samples at phase 0 (co-sited) with respect to the high-res grid
 * \returns void
 */
void downScaleComponent2x2ZeroPhase( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight,const int filt_type){

  const int iDestWidth = iSrcWidth/2;
  const int iDestHeight = iSrcHeight/2;

  const int HALF_NUM_DOWN_TAPS = dn_filt_half_lens0[filt_type];
  const int* iTaps = dn_filt_taps0[filt_type];
  const int iFilterShift = dn_filt_shifts0[filt_type];
  const int iCentreTap = 1<<(iFilterShift-1);

  const int iTempStride = iSrcWidth + 2*HALF_NUM_DOWN_TAPS;
  Pel* pTemp = new Pel[ iDestHeight * iTempStride];

  int iTemp;

  // Vertical
  const Pel* pDataSrc = pSrc;
  Pel* pDataDest = pTemp + HALF_NUM_DOWN_TAPS;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0; i<iSrcWidth; ++i){
      iTemp = pDataSrc[i]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[i+(k+1)*iSrcStride] + pDataSrc[i-(k+1)*iSrcStride]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);
    }
    std::fill(pDataDest - HALF_NUM_DOWN_TAPS, pDataDest, pDataDest[0] );
    std::fill(pDataDest + iSrcWidth, pDataDest+iSrcWidth+HALF_NUM_DOWN_TAPS, pDataDest[iSrcWidth-1] );
    pDataSrc += 2*iSrcStride;
    pDataDest += iTempStride;
  }

  // Horizontal
  pDataSrc = pTemp+HALF_NUM_DOWN_TAPS;
  pDataDest = pDest;

  for (int j=0; j<iDestHeight; ++j){
    for (int i=0,ii=0; i<iDestWidth; ++i,ii+=2){
      iTemp = pDataSrc[ii]*iCentreTap;
      for (int k=0; k<HALF_NUM_DOWN_TAPS; ++k){
        iTemp += iTaps[k]*(pDataSrc[ii+k+1] + pDataSrc[ii-k-1]);
      }
      iTemp += 1<<(iFilterShift-1);
      pDataDest[i] = Clip(iTemp>>iFilterShift);

    }
    pDataSrc += iTempStride;
    pDataDest += iDestStride;
  }

  delete[] pTemp;

}

/** Non-normative downscaling of a picture by 2x2, e.g. for original input pictures
 * \returns void
 */
void downScalePic(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==0 && iDestResIdx == 1){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    downScaleComponent2x2ZeroPhase( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight(), 1);
    downScaleComponent2x2ZeroPhase( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 1 );
    downScaleComponent2x2ZeroPhase( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 1 );
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}


/** Normative downscaling of a reference picture by 2x2
 * \returns void
 */
void downScaleReference(const Int iSrcResIdx, const Int iDestResIdx, TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest){

  if (iSrcResIdx==0 && iDestResIdx == 1){
    pcYuvSrc->setBorderExtension(false);
    pcYuvSrc->extendPicBorder();
    if (REF_DN_FILT==0) {
      downScaleComponent2x2ZeroPhase( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight(), 0);
      downScaleComponent2x2ZeroPhase( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 0 );
      downScaleComponent2x2ZeroPhase( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2, 0 );
    } else {
      downScaleComponent2x2HalfPhase( pcYuvSrc->getLumaAddr(), pcYuvDest->getLumaAddr(), pcYuvSrc->getStride(), pcYuvDest->getStride(), pcYuvSrc->getWidth(), pcYuvSrc->getHeight());
      downScaleComponent2x2HalfPhase( pcYuvSrc->getCbAddr(), pcYuvDest->getCbAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2);
      downScaleComponent2x2HalfPhase( pcYuvSrc->getCrAddr(), pcYuvDest->getCrAddr(), pcYuvSrc->getCStride(), pcYuvDest->getCStride(), pcYuvSrc->getWidth()/2, pcYuvSrc->getHeight()/2);
    }
    pcYuvDest->setBorderExtension(false);
    pcYuvDest->extendPicBorder();
  }
}

//! \}

